"""
pytest fixtures for unit testing
"""

import pytest
from whendo.core.dispatcher import Dispatcher
from whendo.core.actions.file_action import FileHeartbeat
from whendo.core.scheduler import Timely
from whendo.core.continuous import Continuous
from whendo.api import main_temp
from .uvicorn_server import UvicornTestServer


@pytest.fixture
def friends(tmp_path):
    """ returns a tuple of useful test objects """

    def stuff():
        # want a fresh tuple from the fixture
        saved_dir = str(tmp_path)
        output_file = str(tmp_path / "output.txt")
        dispatcher = Dispatcher(saved_dir=saved_dir)
        dispatcher.set_continuous(Continuous())
        dispatcher.initialize()
        action = FileHeartbeat(file=output_file)
        scheduler = Timely(interval=1)

        return dispatcher, scheduler, action

    return stuff


@pytest.fixture
def port(unused_tcp_port):
    """ returns an unused tcp port """
    return unused_tcp_port


@pytest.fixture
def host():
    """ the host used in unit tests """
    return "127.0.0.1"


@pytest.fixture
async def startup_and_shutdown_uvicorn(host, port):
    """Start server as test fixture and tear down after test"""
    uvicorn_server = UvicornTestServer(app=main_temp.app, host=host, port=port)
    await uvicorn_server.up()
    yield
    await uvicorn_server.down()


@pytest.fixture
def base_url(host, port):
    """ base url for unit test invocation of api """
    return f"http://{host}:{port}"



from fastapi import FastAPI
import asyncio
import uvicorn
from typing import Optional, List
from whendo.core.util import SystemInfo


class UvicornTestServer(uvicorn.Server):
    """Uvicorn test server

    This class is slight adaptation of the result of the discussion in this github thread...

        https://github.com/miguelgrinberg/python-socketio/issues/332#issuecomment-712928157

    Usage:
        @pytest.fixture
        async def start_stop_server():
            server = UvicornTestServer()
            await server.up()
            yield
            await server.down()

    Note:
        The fixture mechanism calls the above fixture twice. Once at the beginning of the
        relevant test function, and once at the exit of the test function. The yield
        basically splits the execution into two parts: setup and teardown. The fixtures
        package and the people who used fixtures to build this class are awesome.
    """

    def __init__(self, app: FastAPI, host: str, port: int, workers=1):
        """
        Create a Uvicorn test server for use as a pytest fixture
        """
        SystemInfo.init(host=host, port=port)
        self._startup_done = asyncio.Event()
        super().__init__(config=uvicorn.Config(app, host=host, port=port, workers=1))

    async def startup(self, sockets: Optional[List] = None) -> None:
        """Override uvicorn startup"""
        await super().startup(sockets=sockets)
        self.config.setup_event_loop()
        self._startup_done.set()

    async def up(self) -> None:
        """Start up server asynchronously"""
        self._serve_task = asyncio.create_task(self.serve())
        await self._startup_done.wait()

    async def down(self) -> None:
        """Shut down server asynchronously"""
        self.should_exit = True
        await self._serve_task



from pydantic import BaseModel
from httpx import AsyncClient
from whendo.core.action import Action
from whendo.core.scheduler import Scheduler
from whendo.core.resolver import resolve_action, resolve_scheduler, resolve_file_pathe
from whendo.core.util import FilePathe, DateTime
from whendo.core.dispatcher import Dispatcher


class ClientAsync(BaseModel):
    """
    This class is very similar to Client, except that the methods
    that typically return BaseModel objects will return supplied
    dictionaries if resolution fails.
    """

    host: str = "127.0.0.1"
    port: int = 8000

    # /dispatcher
    async def load_dispatcher(self):
        return Dispatcher.resolve(await self.get_as_json(f"/dispatcher/load"))

    async def save_dispatcher(self):
        return await self.get("/dispatcher/save")

    async def clear_dispatcher(self):
        return await self.get("/dispatcher/clear")

    async def load_dispatcher_from_name(self, name: str):
        return Dispatcher.resolve(
            await self.get_as_json(f"/dispatcher/load_from_name/{name}")
        )

    async def save_data_to_name(self, name: str):
        return await self.get(f"/dispatcher/save_to_name/{name}")

    async def get_saved_dir(self):
        return resolve_file_pathe(
            await self.get_as_json("/dispatcher/saved_dir"), check_for_found_class=False
        )

    async def set_saved_dir(self, saved_dir: FilePathe):
        return await self.put("/dispatcher/saved_dir", saved_dir)

    async def replace_dispatcher(self, replacement: Dispatcher):
        return await self.put("/dispatcher/replace", replacement)

    # /execution
    async def execute_supplied_action(self, supplied_action: Action):
        return await self.put("/execution", supplied_action)

    # /actions
    async def get_action(self, action_name: str):
        return resolve_action(
            await self.get_as_json(f"/actions/{action_name}"),
            check_for_found_class=False,
        )

    async def add_action(self, action_name: str, action: Action):
        return await self.post(f"/actions/{action_name}", action)

    async def set_action(self, action_name: str, action: Action):
        return await self.put(f"/actions/{action_name}", action)

    async def delete_action(self, action_name: str):
        return await self.delete(f"/actions/{action_name}")

    async def execute_action(self, action_name: str):
        return await self.get(f"/actions/{action_name}/execute")

    async def reschedule_action(self, action_name: str):
        return await self.get(f"/actions/{action_name}/reschedule")

    async def unschedule_action(self, action_name: str):
        return await self.get(f"/actions/{action_name}/unschedule")

    # /schedulers

    async def schedule_action(self, scheduler_name: str, action_name: str):
        return await self.get(f"/schedulers/{scheduler_name}/actions/{action_name}")

    async def get_scheduler(self, scheduler_name: str):
        return resolve_scheduler(
            await self.get_as_json(f"/schedulers/{scheduler_name}"),
            check_for_found_class=False,
        )

    async def add_scheduler(self, scheduler_name: str, scheduler: Scheduler):
        return await self.post(f"/schedulers/{scheduler_name}", scheduler)

    async def set_scheduler(self, scheduler_name: str, scheduler: Scheduler):
        return await self.put(f"/schedulers/{scheduler_name}", scheduler)

    async def delete_scheduler(self, scheduler_name: str):
        return await self.delete(f"/schedulers/{scheduler_name}")

    async def unschedule_scheduler(self, scheduler_name: str):
        return await self.get(f"/schedulers/{scheduler_name}/unschedule")

    async def reschedule_all_schedulers(self):
        return await self.get(f"/schedulers/reschedule_all")

    async def execute_scheduler_actions(self, scheduler_name: str):
        return await self.get(f"/schedulers/{scheduler_name}/execute")

    async def scheduled_action_count(self):
        return await self.get("/schedulers/action_count")

    # deferrals and expirations
    async def defer_action(
        self, scheduler_name: str, action_name: str, wait_until: DateTime
    ):
        return await self.post(
            f"/schedulers/{scheduler_name}/actions/{action_name}/defer", wait_until
        )

    async def clear_deferred_actions(self):
        return await self.get(f"/schedulers/clear_deferred_actions")

    async def deferred_action_count(self):
        return await self.get("/schedulers/deferred_action_count")

    async def expire_action(
        self, scheduler_name: str, action_name: str, expire_on: DateTime
    ):
        return await self.post(
            f"/schedulers/{scheduler_name}/actions/{action_name}/expire", expire_on
        )

    async def clear_expired_actions(self):
        return await self.get(f"/schedulers/clear_expired_actions")

    async def expired_action_count(self):
        return await self.get("/schedulers/expired_action_count")

    # /jobs
    async def run_jobs(self):
        return await self.get(f"/jobs/run")

    async def stop_jobs(self):
        return await self.get(f"/jobs/stop")

    async def jobs_are_running(self):
        return await self.get(f"/jobs/are_running")

    async def job_count(self):
        return await self.get(f"/jobs/count")

    async def clear_jobs(self):
        return await self.get(f"/jobs/clear")

    # verbs
    async def get(self, path: str):
        async with AsyncClient(base_url=self.base_url()) as ac:
            return await ac.get(url=path)

    async def get_as_json(self, path: str):
        async with AsyncClient(base_url=self.base_url()) as ac:
            response = await ac.get(url=path)
            return response.json()

    async def put(self, path: str, data: BaseModel):
        async with AsyncClient(base_url=self.base_url()) as ac:
            return await ac.put(url=path, data=data.json())

    async def post(self, path: str, data: BaseModel):
        async with AsyncClient(base_url=self.base_url()) as ac:
            return await ac.post(url=path, data=data.json())

    async def delete(self, path: str):
        async with AsyncClient(base_url=self.base_url()) as ac:
            return await ac.delete(url=path)

    def base_url(self):
        return f"http://{self.host}:{self.port}"



from fastapi import APIRouter, status, Depends
from whendo.api.shared import return_success, raised_exception, get_dispatcher
from whendo.core.dispatcher import Dispatcher
from whendo.core.util import FilePathe

router = APIRouter(prefix="/dispatcher", tags=["Dispatcher"])


@router.get("/clear", status_code=status.HTTP_200_OK)
def clear():
    try:
        get_dispatcher(router).clear_all()
        return return_success("dispatcher cleared")
    except Exception as e:
        raise raised_exception("failed to clear the Dispatcher", e)


@router.put("/replace", status_code=status.HTTP_200_OK)
def replace(replacement=Depends(Dispatcher.resolve)):
    try:
        assert replacement, f"couldn't resolve class for replacement dispatcher"
        get_dispatcher(router).replace_all(replacement=replacement)
        return return_success(f"dispatcher was successfully replaced")
    except Exception as e:
        raise raised_exception(f"failed to replace dispatcher", e)


@router.get("/load", status_code=status.HTTP_200_OK)
def load():
    try:
        return return_success(get_dispatcher(router).load_current())
    except Exception as e:
        raise raised_exception("failed to retrieve the Dispatcher", e)


@router.get("/save", status_code=status.HTTP_200_OK)
def save():
    try:
        get_dispatcher(router).save_current()
        return return_success(f"dispatcher saved to current")
    except Exception as e:
        raise raised_exception("failed to save the Dispatcher", e)


@router.get("/load_from_name/{name}", status_code=status.HTTP_200_OK)
def load_from_name(name: str):
    try:
        return return_success(get_dispatcher(router).load_from_name(name))
    except Exception as e:
        raise raised_exception(f"failed to retrieve the Dispatcher from ({name})", e)


@router.get("/save_to_name/{name}", status_code=status.HTTP_200_OK)
def save_to_name(name: str):
    try:
        get_dispatcher(router).save_to_name(name)
        return return_success(f"dispatcher saved to ({name})")
    except Exception as e:
        raise raised_exception(f"failed to save the Dispatcher to ({name})", e)


@router.put("/saved_dir", status_code=status.HTTP_200_OK)
def set_saved_dir(file_pathe: FilePathe):
    try:
        saved_dir = file_pathe.path
        get_dispatcher(router).set_saved_dir(saved_dir)
        return return_success(f"saved_dir set to ({saved_dir})")
    except Exception as e:
        raise raised_exception("failed to set (saved_dir)", e)


@router.get("/saved_dir", status_code=status.HTTP_200_OK)
def get_saved_dir():
    try:
        saved_dir = get_dispatcher(router).get_saved_dir()
        file_pathe = FilePathe(path=saved_dir)
        return return_success(file_pathe)
    except Exception as e:
        raise raised_exception("failed to get (saved_dir)", e)



from fastapi import APIRouter, status
from whendo.api.shared import return_success, raised_exception, get_dispatcher

router = APIRouter(prefix="/jobs", tags=["Jobs"])


@router.get("/run", status_code=status.HTTP_200_OK)
def start_scheduled_jobs():
    try:
        assert not get_dispatcher(
            router
        ).jobs_are_running(), "should not be running when trying to start"
        get_dispatcher(router).run_jobs()
        return return_success("started running")
    except Exception as e:
        raise raised_exception("failed to start running", e)


@router.get("/stop", status_code=status.HTTP_200_OK)
def stop_scheduled_jobs():
    try:
        assert get_dispatcher(
            router
        ).jobs_are_running(), "should be running when trying to stop"
        get_dispatcher(router).stop_jobs()
        return return_success("stopped running")
    except Exception as e:
        raise raised_exception("failed to stop running", e)


@router.get("/count", status_code=status.HTTP_200_OK)
def job_count():
    try:
        return return_success({"job_count": get_dispatcher(router).job_count()})
    except Exception as e:
        raise raised_exception("failed to get job count", e)


@router.get("/are_running", status_code=status.HTTP_200_OK)
def jobs_are_running():
    try:
        return return_success(get_dispatcher(router).jobs_are_running())
    except Exception as e:
        raise raised_exception("failed to determine if jobs are running", e)


@router.get("/clear", status_code=status.HTTP_200_OK)
def clear():
    try:
        get_dispatcher(router).clear_jobs()
        return return_success("jobs cleared")
    except Exception as e:
        raise raised_exception("failed to get job count", e)



"""
That these modules are imported here ensures that Action and Scheduler resolution will use their entire
inheritance trees to derive an instance from a dictionary.
"""
import whendo.core.action_imports



from whendo.core.action import Action
from whendo.core.util import SystemInfo


class SystemInfoAction(Action):
    """
    Return system info
    """

    system_info = "system_info"

    def execute(self, tag: str = None, scheduler_info: dict = None):
        return SystemInfo.get()



from pydantic import BaseModel
from whendo.sdk.client import Client


class Workbench(BaseModel):
    clients: dict = {}

    def add_client(self, name: str, client: Client):
        self.clients[name] = client

    def get_client(self, name: str):
        return self.clients.get(name, None)

    def get_clients(self):
        return list(self.clients.values())

    def count(self):
        return len(self.clients)



from typing import List, Dict, Any, Optional
from enum import Enum
import json
from whendo.core.action import Action


class Failure(Action):
    """ acts like False """

    logic_failure:str="failure"

    def execute(self, tag: str = None, scheduler_info: dict = None):
        raise Exception("purposely unsuccessful execution", self.json())


class Success(Action):
    """ acts like True """

    logic_success:str="success"

    def execute(self, tag: str = None, scheduler_info: dict = None):
        return {"outcome": "purposely successful execution", "action": self.info()}


class Not(Action):
    """ acts like negation """

    logic_not:str="not"
    operand: Action

    def execute(self, tag: str = None, scheduler_info: dict = None):
        operand_result = None
        try:
            operand_result = self.operand.execute(
                tag=tag, scheduler_info=scheduler_info
            )
        except Exception as exception:
            operand_result = exception
        if isinstance(operand_result, Exception):
            return {
                "outcome": "exception negated; action execution treated as a success",
                "exception": str(operand_result),
                "action": self.info(),
            }
        else:
            raise Exception(
                "exception generated; action execution treated as a failure",
                self.json(),
            )




class ListOpMode(str, Enum):
    ALL = "all"
    OR = "or"
    AND = "and"
    
class ListAction(Action):
    """
    executes actions based on:

        ListOpMode.ALL:   executes all, irrespective of individual outcomes
        ListOpMode.OR:    executes until the first successful action (no exception)
        ListOpMode.AND:   executes until the first exception (failure)

    """

    op_mode: ListOpMode
    action_list: List[Action]
    exception_on_no_success: bool = False

    def execute(self, tag: str = None, scheduler_info: Dict[str, Any] = None):
        (
            processing_count,
            success_count,
            failure_count,
            successful_actions,
            exception_actions,
        ) = process_action_list(
            tag,
            scheduler_info,
            self.op_mode,
            action_list=self.action_list,
            successful_actions=[],
            exception_actions=[],
        )
        processing_info = {
            "processing_count": processing_count,
            "success_count": success_count,
            "exception_count": failure_count,
            "successful_actions": successful_actions,
            "exception_actions": exception_actions,
        }
        if success_count == 0 and self.exception_on_no_success:
            raise Exception(
                "exception generated; action execution treated as a failure",
                json.dumps(self.dict()),
                json.dumps(processing_info),
            )
        else:
            return {
                "outcome": "list action executed",
                "action": self.info(),
                "processing_info": processing_info,
            }

class All(ListAction):
    """
    executes all actions
    """

    logic_all:str="all"
    op_mode:ListOpMode=ListOpMode.ALL
    def execute(self, tag: str = None, scheduler_info: Dict[str, Any] = None):
        return super().execute(tag=tag, scheduler_info=scheduler_info)

class Or(ListAction):
    """
    executes actions until first success
    """

    logic_or:str="or"
    op_mode:ListOpMode=ListOpMode.OR
    def execute(self, tag: str = None, scheduler_info: Dict[str, Any] = None):
        return super().execute(tag=tag, scheduler_info=scheduler_info)

class And(ListAction):
    """
    executes actions until first failure
    """

    logic_and:str="and"
    op_mode:ListOpMode=ListOpMode.AND
    def execute(self, tag: str = None, scheduler_info: Dict[str, Any] = None):
        return super().execute(tag=tag, scheduler_info=scheduler_info)




class IfElse(Action):
    """
    executes actions based on:

        if test_action is successful, executes if_action
        otherwise executes else_action

    """

    logic_if_else:str="if_else"
    test_action: Action
    if_action: Action
    else_action: Action
    exception_on_no_success: bool = False

    def execute(self, tag: str = None, scheduler_info: dict = None):
        try:
            test_result = self.test_action.execute(
                tag=tag, scheduler_info=scheduler_info
            )
        except Exception as exception:
            test_result = exception
        processing_count = 1
        success_count = 0
        exception_count = 0
        successful_actions = []
        exception_actions = []
        else_test = isinstance(test_result, Exception)
        if else_test:  # execute the else action
            exception_count = 1
            exception_actions.append(self.test_action.dict())

            try:
                else_result = self.else_action.execute(
                    tag=tag, scheduler_info=scheduler_info
                )
            except Exception as exception:
                else_result = exception
            processing_count += 1
            if isinstance(else_result, Exception):
                exception_count += 1
                exception_actions.append(self.else_action.dict())
            else:
                success_count += 1
                successful_actions.append(self.else_action.dict())
        else:
            success_count += 1
            successful_actions.append(self.test_action.dict())

            try:
                if_result = self.if_action.execute(
                    tag=tag, scheduler_info=scheduler_info
                )
            except Exception as exception:
                if_result = exception
            processing_count += 1
            if isinstance(if_result, Exception):
                exception_count += 1
                exception_actions.append(self.if_action.dict())
            else:
                success_count += 1
                successful_actions.append(self.if_action.dict())

        processing_info = {
            "else_test": else_test,
            "processing_count": processing_count,
            "success_count": success_count,
            "exception_count": exception_count,
            "successful_actions": successful_actions,
            "exception_actions": exception_actions,
        }
        if success_count == 0 and self.exception_on_no_success:
            raise Exception(
                "exception generated; action execution treated as a failure",
                json.dumps(self.dict()),
                json.dumps(processing_info),
            )
        else:
            return {
                "outcome": "if else action executed",
                "action": self.info(),
                "processing_info": processing_info,
            }



def process_action_list(
    tag: Optional[str],
    scheduler_info: Optional[Dict[str, Any]],
    op_mode: ListOpMode,
    action_list: List[Action],
    processing_count: int = 0,
    success_count: int = 0,
    exception_count: int = 0,
    successful_actions: List[Dict[Any, Any]] = [],
    exception_actions: List[Dict[Any, Any]] = [],
):
    """
    services the two action list classes
    """
    for action in action_list:
        try:  # in case an Action does not return an exception
            result = action.execute(tag=tag, scheduler_info=scheduler_info)
        except Exception as exception:
            result = exception
        processing_count += 1
        if isinstance(result, Exception):
            exception_count += 1
            exception_actions.append(action.dict())
            if op_mode == ListOpMode.AND:  # stop after first failure
                break
        else:
            success_count += 1
            successful_actions.append(action.dict())
            if op_mode == ListOpMode.OR:  # stop after first success
                break
    return (
        processing_count,
        success_count,
        exception_count,
        successful_actions,
        exception_actions,
    )